//
// 3Cell monitor
//
//                     ---------
//    RS    <-- RA2/AN2|1    18| RA1/AN1 --> RWZ
//  analog1 --> RA3/AN3|2    17| RA0/AN0 --> STB
//  analog2 --> RA4/AN4|3    16| OSC1
//          /-- RA5    |4    15| OSC2
//                  GND|5    14| +5V
//    DB4 <--   RB0    |6    13| RB7/AN6 --> FET3
//    DB5 <--   RB1    |7    12| RB6/AN5 <-- analog3
//    DB6 <--   RB2    |8    11| RB5     --> FET2
//    DB7 <--   RB3    |9    10| RB4     --> FET1
//                     ---------

//#include "16F88.h
//#include P16F88.INC


//#define debug	1


#pragma config  = 0x3F50

#pragma codepage 0
#pragma rambank 0
#pragma update_PAGE 1 /* ON */

#define Bmode 0x40		//port B initial mode  6=input 7,5,4,3,2,1,0=output
#define Amode 0x38		//port A initial mode  7,6,2,1,0=output  5,4,3=input

//-----------------------------------
#define ERROR			0
#define TRUE			1
#define FALSE			0
#define	HIGH			1
#define LOW				0

//-----------------------------------
#pragma	bit rs	@ PORTA.2		//chip select
#pragma bit rw	@ PORTA.1		//read/write
#pragma bit stb	@ PORTA.0		//strobe

#pragma bit FET3 @ PORTB.7		//FET3 drive port
#pragma bit FET2 @ PORTB.5		//FET2 drive port
#pragma bit FET1 @ PORTB.4		//FET1 drive port

//-----------------------------------
#pragma bit RA0 @ PORTA.0
#pragma bit RA1 @ PORTA.1
#pragma bit RA2 @ PORTA.2
#pragma bit RA3 @ PORTA.3
#pragma bit RA4 @ PORTA.4
#pragma bit RA5 @ PORTA.5
#pragma bit RA6 @ PORTA.6

#pragma bit RB0 @ PORTB.0
#pragma bit RB1 @ PORTB.1
#pragma bit RB2 @ PORTB.2
#pragma bit RB3 @ PORTB.3
#pragma bit RB4 @ PORTB.4
#pragma bit RB5 @ PORTB.5



//-----------------------------------------------
int lcd_ready(void);
void lcd_data(void);
void cmdout(int);
void lcd_cmd(int);
void lcd_clear(void);
void lcd_incmd(int);
void lcd_init(void);
void cgwrite(int);

extern  page0 void cell3_discharge( void );
extern  page0 void cell2_discharge( void );
extern  page0 void cell1_discharge( void );

extern  page0 void ad5( void );
extern  page0 void ad4( void );
extern  page0 void ad3( void );
extern	page0 void convert2volt(void);
extern	page0 void get_moji(void);

extern	page0 void delay_ms( void );
extern	page0 unsigned int number( unsigned int );


//-----------------------------------------------global
char moji_1[16];
char moji_2[16];
//const page0 char number[]="0123456789";
unsigned int	moji100,moji10,moji1;
unsigned long	ch_v;
unsigned long	delta3, delta2, delta1;

uns16	millisec;


///////////////////////////////////////////////
//  LCD control Library
//  functions are below
//    lcd_init()-------- initialize
//    lcd_ready()------- busy check
//    lcd_cmd(cmd)------ send command
//    lcd_data(string)-- display string
//    lcd_clear() ------ clear display
//////////////////////////////////////////////

/////////// lcd ready check function	
int lcd_ready(void){
	int tmp;
	int high_port,low_port;
	TRISB = Bmode | 0x0F;	//Lower is input
	rs = LOW;
	rw = HIGH;					//read mode
	stb = HIGH;
tmp=0;
	high_port=PORTB & 0x0F;		//input upper
	stb = LOW;
	stb = HIGH;
tmp=0;
	low_port=PORTB & 0x0F;		//input lower
	stb = LOW;
	TRISB = Bmode;
	tmp=high_port<<4;
	tmp=tmp | low_port;
	return(tmp);	//end check
}

////////// lcd display data function
void lcd_data(void){
int tmp,num;

//---------------------------------------line 1
	for(num=0 ; num<=15 ; num++){

		tmp=moji_1[num]>>4;
		tmp=tmp | 0xF0;				//RB[7:4]=HIGH
		PORTB = tmp;				//set upper data
		rw = LOW;					//set write
		rs = HIGH;					//set rs high
		stb = HIGH;					//strobe
		stb = LOW;
	tmp=0;
		tmp=moji_1[num] | 0xF0;		//RB[7:4]=HIGH
		PORTB = tmp;				//set lower data
		stb = HIGH;					//strobe
		stb = LOW;
	tmp=0;

#ifndef debug
		tmp=lcd_ready() & 0x80;	
		while( tmp != 0 ){
			tmp=lcd_ready() & 0x80;	
		}
#endif
	}	// end of for

//---------------------------------------line 2
	lcd_cmd(0xC0);      //DDRAM address second line set

	for(num=0 ; num<=15 ; num++){
		
		tmp=moji_2[num]>>4;
		tmp=tmp | 0xF0;				//RB[7:4]=HIGH
		PORTB = tmp;				//set lower data
		rw = LOW;					//set write
		rs = HIGH;					//set rs high
		stb = HIGH;					//strobe
		stb = LOW;
	tmp=0;
		tmp=moji_2[num] | 0xF0;		//RB[7:4]=HIGH
		PORTB = tmp;				//set upper data
		stb = HIGH;					//strobe
		stb = LOW;
	tmp=0;

#ifndef debug
		tmp=lcd_ready() & 0x80;	
		while( tmp != 0 ){
			tmp=lcd_ready() & 0x80;	
		}
#endif
	}	// end of for

}

////////// lcd command out function
void cmdout(int cmd){
int tmp;
	tmp=cmd>>4;
	tmp=tmp | 0xF0;			//RB[7:4]=HIGH
	PORTB = tmp;			//set upper data
	rw = LOW;				//set write
	rs = LOW;				//set rs low
	stb = HIGH;				//strobe
	stb = LOW;
tmp=0;
	cmd = cmd | 0xF0;		//RB[7:4]=HIGH
	PORTB = cmd;			//set lower data
	stb = HIGH;				//strobe
	stb = LOW;
tmp=0;
}


void lcd_cmd(int cmd){
int tmp;
	cmdout(cmd);
#ifndef debug
	tmp=lcd_ready() & 0x80;	
	while( tmp != 0 ){
		tmp=lcd_ready() & 0x80;	
	}
#endif
}

//////////  lcd display clear function
void lcd_clear(){
	lcd_cmd(1);					//display clear
}

///////// lcd initialize function
void lcd_incmd(int cmd){
int tmp;
	tmp=cmd>>4;
	tmp=tmp | 0xF0;				//RB[7:4]=HIGH
	PORTB = tmp;				//mode command
	rw = LOW;					//set write
	rs = LOW;					//set rs low
	stb = HIGH;					//strobe
	stb = LOW;

	millisec = 1;
	delay_ms();
}

void lcd_init(){
	millisec = 5;
	delay_ms();
	lcd_incmd(0x30);			//8bit mode set
	lcd_incmd(0x30);			//8bit mode set
	lcd_incmd(0x30);			//8bit mode set
	lcd_incmd(0x20);			//4bit mode set
	lcd_cmd(0x2C);				//DL=0 4bit mode  N=1 F=1
	lcd_cmd(0x08);				//display off DI=0 C=0 B=0
	lcd_cmd(0x01);				//display clear
	lcd_cmd(0x06);				//moving mode I/D=1 SH=0
	lcd_cmd(0x0C);				//display on  DI=1 C=0 B=0

	lcd_cmd(0x40);				//CGRAM address 0x0 set
	cgwrite(0x08);				//special character  battery mark
	cgwrite(0x08);
	cgwrite(0x0B);
	cgwrite(0x0B);
	cgwrite(0x0B);
	cgwrite(0x0B);
	cgwrite(0x08);
	cgwrite(0x08);
}


void cgwrite(int cmd){
int tmp;
	tmp = cmd>>4;
	tmp = tmp | 0xF0;		//RB[7:4]=HIGH
	PORTB = tmp;			//set upper data
	rw = LOW;				//set write
	rs = HIGH;				//set rs high
	stb = HIGH;				//strobe
	stb = LOW;
	cmd = cmd | 0xF0;		//RB[7:4]=HIGH
	PORTB = cmd;			//set lower data
	stb = HIGH;				//strobe
	stb = LOW;
#ifndef debug
	tmp=lcd_ready() & 0x80;	
	while( tmp != 0 ){
		tmp=lcd_ready() & 0x80;	
	}
#endif
}



//----------------------------------------------------
void main(void)
{
	unsigned long	v3, v2, v1;
	unsigned long	vmin;


	OSCCON = 0x60;		// internal OSC = 4MHz
	PORTA = 0x00;
	PORTB = 0xB0;		//RB7 RB5 RB4 = HIGH (FET=OFF)
	ANSEL = 0x38;		//AN5,4,3 = A/D 
	ADCON0 = 0xC0;		//
	ADCON1 = 0x00;		//Left justified, ADCS2=1, reference=Vdd/Vss
	CMCON = 0x07;		//Comparator setting Not use
 	OPTION = 0x82;
 	TMR0 = 0;
 	TRISA = Amode;		//#define Amode 0x38   7,6,2,1,0=output  5,4,3=input
 	TRISB = Bmode;		//0x40	port B initial mode  6=input 7,5,4,3,2,1,0=output
	ADON = 1;

//---------------------------------------------
	FET3=HIGH;
	FET2=HIGH;
	FET1=HIGH;

//---------------------------------------------
	millisec = 100;
	delay_ms();     //wait 100msec. MSM6222B is busy after 15msec from power on.

	lcd_init();			//initialize LCD

//RA4=LOW;	//debug
//RA3=HIGH;	//debug
	lcd_clear();       //clear display

	moji_1[0]='3';
	moji_1[1]='C';
	moji_1[2]='e';
	moji_1[3]='l';
	moji_1[4]='l';
	moji_1[5]=' ';
	moji_1[6]='b';
	moji_1[7]='a';
	moji_1[8]='l';
	moji_1[9]='l';
	moji_1[10]='a';
	moji_1[11]='n';
	moji_1[12]='c';
	moji_1[13]='e';
	moji_1[14]='r';
	moji_1[15]=' ';
	
	moji_2[0]=' ';
	moji_2[1]=' ';
	moji_2[2]=' ';
	moji_2[3]=' ';
	moji_2[4]=' ';
	moji_2[5]=' ';
	moji_2[6]=' ';
	moji_2[7]='b';
	moji_2[8]='y';
	moji_2[9]=' ';
	moji_2[10]='n';
	moji_2[11]='a';
	moji_2[12]='r';
	moji_2[13]='i';
	moji_2[14]='a';
	moji_2[15]='i';

	lcd_data();
	millisec = 1000;
	delay_ms();     //wait 1sec

//RA4=HIGH;	//debug
//RA3=LOW;	//debug

	lcd_clear();       //clear display

//                0123456789ABCDEF
//		moji_1[]=" 4.12 4.22 4.31 ";
//		moji_2[]="      |I|I|I GND";
//                 ch5  ch4  ch3

	moji_2[0]=' ';
	moji_2[1]=' ';
	moji_2[2]=' ';
	moji_2[3]=' ';
	moji_2[4]=' ';
	moji_2[5]=' ';
	moji_2[6]=' ';
	moji_2[7]=0;		// battery character
	moji_2[8]=0;		// battery character
	moji_2[9]=0;		// battery character
	moji_2[10]=' ';
	moji_2[11]=' ';
	moji_2[12]='G';
	moji_2[13]='N';
	moji_2[14]='D';
	moji_2[15]=' ';

	lcd_data();
	millisec = 1000;
	delay_ms();     //wait 1sec

	moji_1[4]=' ';
	moji_1[5]=' ';
	moji_1[10]=' ';
	moji_1[11]=' ';

//-------------------------------------------
	while(1){			//endless loop

//------------------------------A/D convert ch5
		ad5();
		convert2volt();
		v3=ch_v;			// 500 > v3 > 0+
		get_moji();

		moji_1[0]=number(moji100);
		moji_1[1]='.';
		moji_1[2]=number(moji10);
		moji_1[3]=number(moji1);
		
//------------------------------A/D convert ch4
		ad4();
		convert2volt();
		v2=ch_v;
		get_moji();

		moji_1[6]=number(moji100);
		moji_1[7]='.';
		moji_1[8]=number(moji10);
			moji_1[9]=number(moji1);

//------------------------------A/D convert ch3
		ad3();
		convert2volt();
		v1=ch_v;
		get_moji();

		moji_1[12]=number(moji100);
		moji_1[13]='.';
			moji_1[14]=number(moji10);
				moji_1[15]=number(moji1);

//--------------------------- LCD display

		lcd_clear();       //clear display
		lcd_data();

//------------------------------decide min Cell voltage

		vmin = v3;			// default

		if( v1<=v2 ){
			if( v1<=v3 ){
				vmin = v1;
			}
		}

		if( v2<=v1 ){
			if( v2<=v3 ){
				vmin = v2;
			}
		}

		if( v3<=v1 ){
			if( v3<=v2 ){
				vmin = v3;
			}
		}

//------------------------------calculate delta voltage from vmin

		delta3 = v3-vmin;
		delta2 = v2-vmin;
		delta1 = v1-vmin;

//--------------------------
		if ( delta3 < 3 ){
			if ( delta2 < 3 ){
				if ( delta1 < 3 ){
					moji_2[0]='B';
					moji_2[1]='a';
					moji_2[2]='l';
					moji_2[3]='l';
					moji_2[4]='a';
					moji_2[5]='n';
					moji_2[6]='c';
					moji_2[7]='e';
					moji_2[8]=' ';
					moji_2[9]='O';
					moji_2[10]='K';
					moji_2[11]=' ';
					moji_2[12]=' ';
					moji_2[13]=' ';
					moji_2[14]=' ';
					moji_2[15]=' ';
					lcd_data();			

					while(1){}				// discharge finish
				}
			}
		}

//--------------------------
		if( delta3 >= 3 ){
			moji_2[0]='C';
			moji_2[1]='e';
			moji_2[2]='l';
			moji_2[3]='l';
			moji_2[4]='-';
			moji_2[5]='3';
			moji_2[6]=' ';
			moji_2[7]='d';
			moji_2[8]='i';
			moji_2[9]='s';
			moji_2[10]='c';
			moji_2[11]='h';
			moji_2[12]='a';
			moji_2[13]='r';
			moji_2[14]='g';
			moji_2[15]='e';
			lcd_data();

			cell3_discharge();

		}

//--------------------------
		if( delta2 >= 3 ){
			moji_2[0]='C';
			moji_2[1]='e';
			moji_2[2]='l';
			moji_2[3]='l';
			moji_2[4]='-';
			moji_2[5]='2';
			moji_2[6]=' ';
			moji_2[7]='d';
			moji_2[8]='i';
			moji_2[9]='s';
			moji_2[10]='c';
			moji_2[11]='h';
			moji_2[12]='a';
			moji_2[13]='r';
			moji_2[14]='g';
			moji_2[15]='e';
			lcd_data();

			cell2_discharge();
		}

//--------------------------
		if( delta1 >= 3 ){
			moji_2[0]='C';
			moji_2[1]='e';
			moji_2[2]='l';
			moji_2[3]='l';
			moji_2[4]='-';
			moji_2[5]='1';
			moji_2[6]=' ';
			moji_2[7]='d';
			moji_2[8]='i';
			moji_2[9]='s';
			moji_2[10]='c';
			moji_2[11]='h';
			moji_2[12]='a';
			moji_2[13]='r';
			moji_2[14]='g';
			moji_2[15]='e';
			lcd_data();

			cell1_discharge();
		}

//--------------------------
		millisec = 1000;
		delay_ms();     //wait 1sec

	}	// end of while
}

